(* qelim.ml *)
(* Contient des fonctions d’interfacage avec Mjollnir. *)

open Formula
open Formula_yices
open Syntax


let plonge f = (f : source_unquantified_formula :> source_quantified_formula)

let varify v =
        match Alpha.get_type v with
        | TBool -> "b" ^ string_of_int v
        | _ -> "v" ^ string_of_int v

let rec convert_to_mjollnir_expr = function

        | Int i, _, _ -> `Integer i
        | Real q, _, _ -> `Quotient (`Integer (Q.get_num q), `Integer (Q.get_den q))

        | Uminus, Op [e], _ -> `Neg (convert_to_mjollnir_expr e)

        | Div, Op [e1 ; e2], _ -> `Quotient (convert_to_mjollnir_expr e1, convert_to_mjollnir_expr e2)
        | Times, Op [e1 ; e2], _ -> `Product [convert_to_mjollnir_expr e1 ; convert_to_mjollnir_expr e2]
        | Plus, Op [e1 ; e2], _ -> `Sum [convert_to_mjollnir_expr e1 ; convert_to_mjollnir_expr e2]
        | Minus, Op [e1 ; e2], _ -> `Sum [convert_to_mjollnir_expr e1 ; `Neg (convert_to_mjollnir_expr e2)]

        | Ident i, _, _ -> `Variable ("v" ^ string_of_int i)

        | _ -> raise Errors.Not_implemented (* FIXME *)

let rec convert_to_mjollnir = function

        | True, _, _ -> formula_true
        | False, _, _ -> formula_false

        | Ident i, _, TBool -> `Atomic (`Comparison (`Variable ("b" ^ string_of_int i), `GREATER, `Integer (Z.zero)))

        | Not, Op [e], _ -> `Not (convert_to_mjollnir e)

        | And, Op [e1 ; e2], _ -> `And [convert_to_mjollnir e1; convert_to_mjollnir e2]
        | Or, Op [e1 ; e2], _ -> `Or [convert_to_mjollnir e1; convert_to_mjollnir e2]
        | Xor, Op [e1 ; e2], _ -> `Xor [convert_to_mjollnir e1; convert_to_mjollnir e2]

        | Impl, Op [e1 ; e2], _ -> `Ite (convert_to_mjollnir e1, convert_to_mjollnir e2, formula_true)

        | Eq, Op [(_, _, TBool) as b1 ; b2], _ -> `Not (`Xor [convert_to_mjollnir b1; convert_to_mjollnir b2])
        | Neq, Op [(_, _, TBool) as b1 ; b2], _ -> `Xor [convert_to_mjollnir b1; convert_to_mjollnir b2]

        | Eq, Op [e1 ; e2], _ -> `Atomic (`Comparison (convert_to_mjollnir_expr e1, `EQUAL, convert_to_mjollnir_expr e2))
        | Neq, Op [e1 ; e2], _ -> `Atomic (`Comparison (convert_to_mjollnir_expr e1, `UNEQUAL, convert_to_mjollnir_expr e2))
        | Gte, Op [e1 ; e2], _ -> `Atomic (`Comparison (convert_to_mjollnir_expr e1, `GREATER_EQ, convert_to_mjollnir_expr e2))
        | Gt, Op [e1 ; e2], _ -> `Atomic (`Comparison (convert_to_mjollnir_expr e1, `GREATER, convert_to_mjollnir_expr e2))
        | Lte, Op [e1 ; e2], _ -> `Atomic (`Comparison (convert_to_mjollnir_expr e1, `LESS_EQ, convert_to_mjollnir_expr e2))
        | Lt, Op [e1 ; e2], _ -> `Atomic (`Comparison (convert_to_mjollnir_expr e1, `LESS, convert_to_mjollnir_expr e2))

        | If, Op [e1 ; e2 ; e3], _ -> `Ite (convert_to_mjollnir e1, convert_to_mjollnir e2, convert_to_mjollnir e3)

        | ((Diese | Nor), _, _) as a -> convert_to_mjollnir (Formules.erase_nor_diese a)

        | (Call _|Real _|Int _|Tuple|Times|Plus|Minus|Mod|Div|Arrow|Real2Int|Int2Real| Uminus), _, _ -> failwith "error while converting to a Mjollnir expression: a unexpected item appear in the expression to be converted."

        | _ -> failwith "problem of aritee while convertir an expression to a Mjollnir expression."

let rec source_expression_to_expression = function

        | `Sum l -> make_sum (List.map source_expression_to_expression l)
        | `Product l -> make_prod (List.map source_expression_to_expression l)
        | `Neg e -> let c, o, t = source_expression_to_expression e in
                    Uminus, Op [c, o, t], t
        | `Quotient (e1, e2) -> Div, Op [source_expression_to_expression e1 ; source_expression_to_expression e2], TReal
        | `Integer i -> from_int i
        | `Variable name ->
                if String.length name = 0 then failwith "Internal error while converting expression from Mjollnir.";
                let id = int_of_string (String.sub name 1 (String.length name - 1)) in
                Alpha.var_ident id

let rec convert_to_expression = function

        | `And l -> make_and (List.map convert_to_expression l)
        | `Or l -> make_or (List.map convert_to_expression l)
        | `Xor l -> make_xor (List.map convert_to_expression l)
        | `Not f -> Not, Op [convert_to_expression f], TBool
        | `Ite (b, t, e) -> If, Op [convert_to_expression b ; convert_to_expression t ; convert_to_expression e], TBool
        | `Atomic (`Propositional b) -> Ident (int_of_string b), Op [], TBool
        | `Atomic (`Comparison (e1, c, e2)) ->
            let ex1 = source_expression_to_expression e1 in
            let ex2 = source_expression_to_expression e2 in
            match ex1, ex2 with
            | (b, Op l, TBool), (Int z, _, _) when z = Z.zero ->
                     (match c, b, l with
                     | (`GREATER | `GREATER_EQ), Ident i, _
                     | (`LESS | `LESS_EQ), Uminus, [Ident i, _, _]
                        -> Ident i, Op [], TBool
                     | (`GREATER | `GREATER_EQ), Uminus, [Ident i, _, _]
                     | (`LESS | `LESS_EQ), Ident i, _
                        -> Not, Op [Ident i, Op [], TBool], TBool
                     | (`EQUAL | `UNEQUAL), _, _ -> failwith "Unexpected output from Mjollnir."
                     | _ -> failwith "Error while reading output from Mjollnir."
                     )
            | _ ->
                    (match c with
                    | `LESS -> Lt
                    | `GREATER -> Gt
                    | `LESS_EQ -> Lte
                    | `GREATER_EQ -> Gte
                    | `EQUAL -> Eq
                    | `UNEQUAL -> Neq
                    ), Op [ex1 ; ex2], TBool

let rec exists l f =
        match l with
        | [] -> f
        | v :: l' ->
                exists l' (mk_exists (varify v) f)

let elim qf =
        reduce formula_true qf

let string_of_var name =
        let id = int_of_string (String.sub name 1 (String.length name - 1)) in
        name ^ "[" ^ Alpha.get_name id ^ "]"

let rec string_of_mjollnir_expr = function
        | `Sum [] -> "0"
        | `Sum [a] -> string_of_mjollnir_expr a
        | `Sum (a :: l) -> "(" ^ string_of_mjollnir_expr a ^ ") + " ^ string_of_mjollnir_expr (`Sum l)
        | `Product [] -> "1"
        | `Product [a] -> string_of_mjollnir_expr a
        | `Product (a :: l) -> "(" ^ string_of_mjollnir_expr a ^ ") * " ^ string_of_mjollnir_expr (`Product l)
        | `Neg e -> "not " ^ string_of_mjollnir_expr e
        | `Quotient (e1, e2) -> "(" ^ string_of_mjollnir_expr e1 ^ ") / (" ^ string_of_mjollnir_expr e2 ^ ")"
        | `Integer i -> Z.to_string i
        | `Variable name -> string_of_var name

let string_of_comparison = function
        | `GREATER -> ">"
        | `GREATER_EQ -> ">="
        | `LESS -> "<"
        | `LESS_EQ -> "<="
        | `EQUAL -> "="
        | `UNEQUAL -> "!="

let string_of_atomic = function
        | `Propositional b -> "prop(" ^ b ^ " [" ^ Alpha.get_name (int_of_string b) ^ "])"
        | `Comparison (e1, c, e2) -> string_of_mjollnir_expr e1 ^ " " ^ string_of_comparison c ^ " " ^ string_of_mjollnir_expr e2

let rec string_of_mjollnir_prop = function
        | `And [] -> "True"
        | (`Or [] | `Xor []) -> "False"
        | (`And [a] | `Or [a] | `Xor [a]) -> string_of_mjollnir_prop a
        | `And (a :: l) -> string_of_mjollnir_prop a ^ " and " ^ string_of_mjollnir_prop (`And l)
        | `Or (a :: l) -> string_of_mjollnir_prop a ^ " or " ^ string_of_mjollnir_prop (`Or l)
        | `Xor (a :: l) -> string_of_mjollnir_prop a ^ " xor " ^ string_of_mjollnir_prop (`Xor l)
        | `Not f -> "not " ^ string_of_mjollnir_prop f
        | `Ite (b, t, e) -> "If " ^ string_of_mjollnir_prop b ^ " then " ^ string_of_mjollnir_prop t ^ " else " ^ string_of_mjollnir_prop e
        | `Atomic a -> string_of_atomic a
        | `Exists _
        | `Forall _ -> raise Errors.Not_implemented

let print_atomic out atomic = output_string out (string_of_atomic atomic)
let print_variable out var = output_string out (string_of_var var)

