(* errors.ml *)
(* Gère les messages d’erreurs. *)

exception Illegal_option of string
exception Help

exception Not_implemented

exception Const_forbidden of Lexeme.t
exception Typedef_forbidden of Lexeme.t
exception Clock_forbidden of Lexeme.t
exception NamedType_forbidden of Lexeme.t
exception Array_forbidden of Lexeme.t
exception Field_forbidden of Lexeme.t
exception CurrentWhen_forbidden of Lexeme.t
exception Case_forbidden of Lexeme.t

exception Aritee_failed of (int * int * Lexeme.t)

exception Name_not_in_table of string

exception Yices_unknown
exception Yices_no_solution


let usage =
	String.concat "\n\t" ["Usage: ";
        Sys.executable_name ^ " [options] [<input file>]";
        "";
        "-h, --help\t\t\tPrint this help and exit";
        "-i, --input <file>\t\tRead input from file";
        "-o, --output <file>\t\tWrite output to file";
        "-v, --verbose\t\t\tVerbose mode";
        "-n, --node <node>\t\tSpecify a specific node to be analysed";
        "-lus, --lustre\t\t\tPrint a Lustre code as an output";
        "-pred, --predicates\t\tPrint the list of predicates used by the algorithm";
        "-pl, --predicates-level <level>\tSet the research of predicates level (from 0 to 4)";
        "-st, --simplify-trivials\tSimplify trivial expressions in the Lustre code";
        "-sif, --simplify-if\t\tSeparate if blocs of the other variables in the Lustre code";
        "-ep, --expend-pre\t\tExpand all pre blocs in the code";
        "-lo, --logic-node\t\tPrint the logical formulae of transition and initialisation of a node";
        "-inv, --find-invariant <n>\tCompute and print an invariant with n clauses of the selected node";
        "\t\t\t\tSet <n> to auto to let the program find a good number of clauses";
        "-vv, --verbose-invariant\tVery verbose mode while computing an invariant property";
        "-lvv, --just-little-verbose\tJust print the number of clauses of founded invariants";
        "-invm, --invariant-not-minimal\tPrint an inductive invariant whithout computing a minimum one";
        "-invkp, --let-permutations\tDo not perform the Removal of Permutations optimisation";
        "-invrsd, --remove-subsumed\tPerform the Removal of Subsumed Disjuncts optimisation";
        "-invsat, --check-sat\t\tImpose the clauses to be satisfiable";
        "-invsb, --not-strictly-better\tWhen searching a better invariant, do not particularly choose one that is mathematically better";
        "-expinv, --expand-invariant\tMake the invariant contains the most possible predicates compatible with it";
        "-nibinv, --not-impose-better\tDo not impose an invariant to be better than the preceding one, but test it later";
        "-aut, --automata <states>\tCompute and print a more or less equivalent automata of the node with the given number of states";
        "\t\t\t\tSet <state> to auto to let the program find a good number of states";
        "-lim, --limit-state <limit>\tLimits the number of state of the automata to the given number"]

let show_usage () = prerr_endline usage

let show_help () = print_endline usage

let we s =
        prerr_endline (Sys.executable_name ^ ": " ^ s)

let show_location t =
        we ("error at line " ^ (string_of_int t.Lexeme.line) ^ ", characters " ^
        (string_of_int t.Lexeme.cstart) ^ "-" ^ (string_of_int t.Lexeme.cend) ^ ".")

let catch = function
        | Failure s ->
                        we ("internal failure: " ^ s)
        | Not_implemented ->
                        we "sorry, not implemented yet."
        | Illegal_option s ->
                        we ("illegal option “" ^ s ^ "”");
                        show_usage ()
        | Const_forbidden t ->
                        we ("constants are forbidden for this program in Lustre files.");
                        show_location t
        | Typedef_forbidden t ->
                        we ("type definition are forbidden for this program in Lustre files.");
                        show_location t
        | Clock_forbidden t ->
                        we ("clocks are forbidden for this program in Lustre files.");
                        show_location t
        | NamedType_forbidden t ->
                        we ("new types are forbidden for this program in Lustre files.");
                        show_location t
        | Array_forbidden t ->
                        we ("arrays are forbidden for this program in Lustre files.");
                        show_location t
        | Field_forbidden t ->
                        we ("fields are forbidden for this program in Lustre files.");
                        show_location t
        | CurrentWhen_forbidden t ->
                        we ("when and current operators are forbidden for this program in Lustre files.");
                        show_location t
        | Case_forbidden t ->
                        we ("test case operators are forbidden for this program in Lustre files.");
                        show_location t
        | Aritee_failed (n, l, t) ->
                        we ("aritee of an operator non respected: should be " ^
                        string_of_int n ^ "rather than" ^ string_of_int l ^ ".");
                        show_location t
        | Name_not_in_table name ->
                        we ("the function or variable “" ^ name ^ "” was requested, but was not declared.");
                        we "maybe the program wasn’t properly and totally loaded."
        | Yices_unknown ->
                        we "Yices returns an unknown statement."
        | Yices_no_solution ->
                        we "no solution for an invariant was found."
        | Help -> show_help ()
        | e -> raise e

