(* lustre.ml *)
(* Contient des fonctions appelant XeC, et parcourant l’arbre pour l’adapter. *)

open Syntaxe
open Hashtbl
open Syntax


let rec convert_type t = function
        | Syntaxe.Bool -> Syntax.TBool
        | Syntaxe.Int -> Syntax.TInt
        | Syntaxe.Real -> Syntax.TReal
        | Syntaxe.TupleType l -> Syntax.TTuple (List.map (convert_type t) l)
        | Syntaxe.NamedType _ -> raise (Errors.NamedType_forbidden t)
        | Syntaxe.ArrayType _ -> raise (Errors.Array_forbidden t)

let create_name s n =
        let rec aux = function
        | 0 -> s
        | n -> "pre_" ^ aux (n - 1)
        in
        let new_name = aux n in
        Alpha.get_new_name new_name

let rec convert_val t l_ref i =
        (* Convertit une valeur de XeC et de Candle. *)
        (* La variable i correspond au nombre de décalages (de « pre » imbriqués
         * autour de la valeur actuelle. *)
        let test_aritee n (Ops l) =
                let longueur = List.length l in
                if n = longueur then true
                else raise (Errors.Aritee_failed (n, longueur, t))
        in
        let mpqf_of_string s = (* Oui, j’ai eu des problèmes avec, et comme le message d’erreur renvoyé par ocamlidl était vide, je préfère le rattraper et faire ce que je peux. *)
                try Mpqf.of_string s with _ ->
                try Mpqf.of_float (float_of_string s) with _ ->
                failwith ("Unable to convert string “" ^ s ^ "” to an Mpqf number.")
        in function
        (* zéroaires *)
        | TRUE_n, args -> let _ = test_aritee 0 args in True, Op [], TBool
        | FALSE_n, args -> let _ = test_aritee 0 args in False, Op [], TBool
        | ICONST_n s, args -> let _ = test_aritee 0 args in Int (Mpzf.of_string s), Op [], TInt
        | RCONST_n s, args -> let _ = test_aritee 0 args in Real (mpqf_of_string s), Op [], TReal
        | (IVAL_n _ | RVAL_n _), args -> let _ = test_aritee 0 args in failwith "IVAL_n or RVAL_n in an expression generated by XeC."
        | IDENT_n s, args ->
                        let _ = test_aritee 0 args in
                        let var = (Alpha.get_ident s) in
                        if i = 0 then Alpha.var_ident var
                        else
                        begin
                                let rec search_in_l = function
                                        | (id, a, j) :: _ when i = j && a = var -> id
                                        | _ :: l -> search_in_l l
                                        | [] -> (* Il faut donc allouer une nouvelle variable. *)
                                                let ty = Alpha.get_type var
                                                in
                                                let new_id = Alpha.get_new_id
                                                (create_name s i) ((*Some*) ty) in
                                                l_ref := (new_id, var, i) :: !l_ref;
                                                new_id
                                in Alpha.var_ident (search_in_l !l_ref)
                        end
        (* unaires *)
        | PRE_n, Ops [e] -> convert_val t l_ref (succ i) e
        | NOT_n, Ops [e] -> Not, Op [convert_val t l_ref i e], TBool
        | UMINUS_n, Ops [e] -> Uminus, Op [convert_val t l_ref i e], TInt
        | INT2REAL_n, Ops [e] -> Int2Real, Op [convert_val t l_ref i e], TReal
        | REAL2INT_n, Ops [e] -> Real2Int, Op [convert_val t l_ref i e], TInt
        | (PRE_n | NOT_n | UMINUS_n | INT2REAL_n | REAL2INT_n), Ops l ->
                        raise (Errors.Aritee_failed (1, List.length l, t))
        | CURRENT_n, _ -> raise (Errors.CurrentWhen_forbidden t)
        (* binaires *)
        | ARROW_n, Ops [e1; e2] -> let v1, v2 = convert_val t l_ref i e1, convert_val t l_ref i e2 in let _, _, ty = v1 in Arrow, Op [v1 ; v2], ty
        | AND_n, Ops [e1; e2] -> And, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | OR_n, Ops [e1; e2] -> Or, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | XOR_n, Ops [e1; e2] -> Xor, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | IMPL_n, Ops [e1; e2] -> Impl, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | EQ_n, Ops [e1; e2] -> Eq, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | NEQ_n, Ops [e1; e2] -> Neq, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | GTE_n, Ops [e1; e2] -> Gte, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | GT_n, Ops [e1; e2] -> Gt, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | LTE_n, Ops [e1; e2] -> Lte, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | LT_n, Ops [e1; e2] -> Lt, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | (DIV_n | SLASH_n), Ops [e1; e2] -> Div, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TBool
        | MOD_n, Ops [e1; e2] -> Mod, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TInt
        | MINUS_n, Ops [e1; e2] -> Minus, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TInt
        | PLUS_n, Ops [e1; e2] -> Plus, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TInt
        | TIMES_n, Ops [e1; e2] -> Times, Op [convert_val t l_ref i e1; convert_val t l_ref i e2], TInt
        | HAT_n, _ -> raise (Errors.Array_forbidden t)
        | WHEN_n, _ -> raise (Errors.CurrentWhen_forbidden t)
        | (ARROW_n | AND_n | OR_n | XOR_n | IMPL_n | EQ_n | NEQ_n | GTE_n | GT_n | LTE_n | LT_n | DIV_n | MOD_n | MINUS_n | PLUS_n | SLASH_n | TIMES_n), Ops l ->
                        raise (Errors.Aritee_failed (2, List.length l, t))
        (* Ternaires *)
        | IF_n, Ops [e1; e2; e3] -> let v1, v2, v3 = convert_val t l_ref i e1, convert_val t l_ref i e2, convert_val t l_ref i e3 in let _, _, ty = v2 in If, Op [v1 ; v2 ; v3], ty
        | IF_n, Ops l -> raise (Errors.Aritee_failed (3, List.length l, t))
        (* n-aires *)
        | NOR_n, Ops l -> Nor, Op (List.map (convert_val t l_ref i) l), TBool
        | DIESE_n, Ops l -> Diese, Op (List.map (convert_val t l_ref i) l), TBool
        | TUPLE_n, Ops l -> let vl = List.map (convert_val t l_ref i) l in Tuple, Op vl, TTuple (List.map (fun (_, _, ty) -> ty) vl)
        | ARRAY_n, _ -> raise (Errors.Array_forbidden t)
        (* Autres *)
        | CALL_n s, Ops l ->
                        let id = Alpha.get_ident s in Call id, Op (List.map (convert_val t l_ref i) l), Alpha.get_type id
        | (FROM_n _ | TO_n _ | ARRAY_ACCES_n _ | ARRAY_SLICE_n _), _ ->
                        raise (Errors.Array_forbidden t)
        | STRUCT_ACCESS_n _, _ ->
                        raise (Errors.Field_forbidden t)
        | CASE_INT_n _, _ ->
                        raise (Errors.Case_forbidden t)
        | SELECT_n _, _ ->
                        failwith "a SELECT block shouldn’t appear in the result of XeC!"

let rec convert_left_part t = function
        | [] -> []
        | LeftVar s :: l -> Alpha.get_ident s :: convert_left_part t l
        | LeftField _ :: _ -> raise (Errors.Field_forbidden t)
        | (LeftArray _ | LeftSlice _) :: _ -> raise (Errors.Array_forbidden t)

let change_equation t l_ref = function
        (* Elle va convertir l’arbre de l’équation de façon à ne plus avoir de
         * « pre » dans les expressions. *)
        | (l, expr) -> (convert_left_part t l, convert_val t l_ref 0 expr)

let id_of_flow_decl t = function
        | (_, _, NamedClock _) -> raise (Errors.Clock_forbidden t)
        | (name, ty, BaseClock) ->
                        Alpha.get_new_id name (convert_type t ty)

let node_of_node_info t node_info =
        let ret = {
                id = -1;

                input = List.map (id_of_flow_decl t) node_info.ni_inputs;
                temporary = List.map (id_of_flow_decl t) node_info.ni_locals;
                output = List.map (id_of_flow_decl t) node_info.ni_outputs;
                memory = [];

                assume = -1;

                equations = [];
        } in
        ret.id <- Alpha.get_new_id node_info.ni_name (match ret.output with
        | [o] -> Alpha.get_type o
        | l -> TTuple (List.map Alpha.get_type l)
        );
        let memory_var = ref [] in (* Va contenir des triplets
        (identifiant_de_la_nouvelle_variable,
        identifiant_de_la_variable_suivie, nb_de_décalages) *)
        ret.equations <- List.map (change_equation t memory_var) node_info.ni_defs;
        let id_assume = Alpha.get_new_id (node_info.ni_name ^ "_assume") TBool in
        let id_pre_assume = Alpha.get_new_id ("pre_" ^ Alpha.get_name id_assume) TBool in
        ret.assume <- id_assume;
        ret.temporary <- id_assume :: ret.temporary;
        let eq_assert =
                [id_assume],
                (Arrow, Op [from_bool true ; (make_and (Alpha.var_ident id_pre_assume :: List.map (convert_val t memory_var 0) node_info.ni_asserts))], TBool)
        in
        ret.equations <- eq_assert :: ret.equations;
        ret.memory <- !memory_var;
        ret.memory <- (id_pre_assume, id_assume, 1) :: ret.memory;
        ret

let load_program in_channel =
        let pack = Parsers.read_xec in_channel in
        function
        | None -> (* On charge tous les nodes *)
        Utils.remove_option (List.map (function
                | ConstDef c ->
                       raise (Errors.Const_forbidden (find pack.const_table c).Lexeme.src)
                | TypeDef t ->
                       raise (Errors.Typedef_forbidden (find pack.const_table t).Lexeme.src)
                | OperDef s ->
                       let oper = find pack.oper_table s in
                       match oper.Lexeme.it with
                       | FuncOper f ->
                                       let automata = Automata.trivial_automata () in
                                       automata.Automata.input <- List.map (fun (n, t) -> Alpha.get_new_id n (convert_type oper.Lexeme.src t)) f.fi_inpars;
                                       automata.Automata.output <- List.map (fun (n, t) -> Alpha.get_new_id n (convert_type oper.Lexeme.src t)) f.fi_outpars;
                                       let id = Alpha.get_new_id f.fi_name (match automata.Automata.output with
                                       | [o] -> Alpha.get_type o
                                       | l -> TTuple (List.map Alpha.get_type l)
                                       ) in
                                       Automata.give_automata id automata;
                                       None
                       | NodeOper n ->  Some (node_of_node_info oper.Lexeme.src n)
        ) pack.def_list)
        | Some node -> (* On nous a donné le nom d’un nœud particulier : on peut donc se limiter à son analyse, sans analyser tout le programme. *)
        let n = try (find pack.oper_table node) with
        | Not_found -> failwith ("no node named “" ^ node ^ "” found in the program.")
        in
        match n.Lexeme.it with
        | FuncOper _ ->
                        failwith ("“" ^ node ^ "” is not the name of a node, but of a external function: it can’t be analysed.")
        | NodeOper ni ->
                        node_of_node_info n.Lexeme.src ni :: []

