(* main.ml *)
(* Filtre les arguments du programme et le lance. *)

open Flux
open Node_action


let rec parse arg =
        let need_predicates () = arg.compute_predicates <- true in
        let need_simplify_if () = arg.simplify_if_blocs <- true in
        let need_expand_pre () = arg.expand_pre_blocs <- true in
        let need_logic_node () = need_simplify_if (); need_expand_pre (); arg.compute_logic_node <- true in
        let need_invariant b i = need_predicates (); need_logic_node ();
        arg.find_invariant <- true; arg.invariant_options.Invariants.size <- i; if i <= 0 then failwith "the requested number of clauses must be greater or equal to one.";
        arg.invariant_options.Invariants.search_better_size <- b in
        let need_automata b i = need_invariant b i; arg.invariant_options.Invariants.generate_automata <- true; arg.compute_automata <- true in
        function
        | [] -> ()
        | ("-h" | "--help") :: _ ->
                        raise Errors.Help
        | ("-n" | "--node") :: node :: s ->
                        arg.node_to_be_analysed <- Some node; parse arg s
        | ("-lus" | "--lustre") :: s ->
                        arg.write_lustre <- true; parse arg s
        | ("-pred" | "--predicates") :: s ->
                        need_predicates (); arg.print_predicates <- true; parse arg s
        | ("-pl" | "--predicates-level") :: i :: s ->
                        need_predicates (); arg.level_predicates <- int_of_string i; parse arg s
        | ("-lo" | "--logic-node") :: s ->
                        need_logic_node (); arg.print_logic_node <- true; parse arg s
        | ("-st" | "--simplify-trivials") :: s ->
                        arg.simplify_trivial_equation <- true; parse arg s
        | ("-sif" | "--simplify-if") :: s ->
                        need_simplify_if (); parse arg s
        | ("-ep" | "--expand-pre") :: s ->
                        need_expand_pre (); parse arg s
        | ("-inv" | "--find-invariant") :: "auto" :: s ->
                        need_invariant true 1; arg.print_invariant <- true; parse arg s
        | ("-inv" | "--find-invariant") :: i :: s ->
                        need_invariant false (int_of_string i); arg.print_invariant <- true; parse arg s
        | ("-vv" | "--verbose-invariant") :: s ->
                        arg.invariant_options.Invariants.verbose <- true; parse arg s
        | ("-lvv" | "--just-little-verbose") :: s ->
                        arg.invariant_options.Invariants.just_a_little_verbose <- true; parse arg s
        | ("-invm" | "--invariant-not-minimal") :: s ->
                        arg.invariant_options.Invariants.minimal <- false; parse arg s
        | ("-invkp" | "--let-permutations") :: s ->
                        arg.invariant_options.Invariants.removal_of_permutation <- false; parse arg s
        | ("-invrsd" | "--remove-subsumed") :: s ->
                        arg.invariant_options.Invariants.removal_of_subsumed_disjuncts <- true; parse arg s
        | ("-invsb" | "--not-strictly-better") :: s ->
                        arg.invariant_options.Invariants.find_better_strictly <- false; parse arg s
        | ("-invsat" | "--check-sat") :: s ->
                        arg.invariant_options.Invariants.satisfiable <- true; parse arg s
        | ("-expinv" | "--expand-invariant") :: s ->
                        arg.invariant_options.Invariants.expand_invariant <- true; parse arg s
        | ("-nibinv" | "--not-impose-better") :: s ->
                        arg.invariant_options.Invariants.variante_eviter_seg_fault_yices <- true; parse arg s
        | ("-aut" | "--automata") :: "auto" :: s ->
                        need_automata true 1; arg.print_automata <- true; parse arg s
        | ("-aut" | "--automata") :: i :: s ->
                        need_automata false (int_of_string i); arg.print_automata <- true; parse arg s
        | ("-lim" | "--limit-state") :: i :: s ->
                        arg.invariant_options.Invariants.max_size <- int_of_string i; parse arg s
        | ("-v" | "--verbose") :: s ->
                        arg.verbose <- true; parse arg s
        | ("-i" | "--input") :: file :: s ->
                        arg.input <- open_in file; parse arg s
        | ("-o" | "--output") :: file :: s ->
                        arg.output <- open_out file; Flux.set_output arg.output; parse arg s
        | file :: [] -> (* Attention : cette clause doit rester en dernière ! *)
                        if file.[0] <> '-' then arg.input <- open_in file
                        else raise (Errors.Illegal_option file)
        | opt :: _ -> raise (Errors.Illegal_option opt)

let parse_arg () =
        let arg =
                {
                        input = stdin;
                        output = stdout;

                        verbose = false;

                        read_lustre = true;
                        write_lustre = false;

                        node_to_be_analysed = None;

                        compute_predicates = false;
                        level_predicates = Logic_node.max_level;
                        print_predicates = false;

                        simplify_trivial_equation = false;
                        simplify_if_blocs = false;
                        expand_pre_blocs = false;

                        compute_logic_node = false;
                        print_logic_node = false;

                        find_invariant = false;
                        invariant_options = Invariants.default;
                        print_invariant = false;

                        compute_automata = false;
                        print_automata = false;
                }
        in
        arg.invariant_options.Invariants.print <- print_string;
        arg.invariant_options.Invariants.print_T <- Printer.print_T;
        parse arg (
                match Array.to_list Sys.argv with
                | [] | _ :: [] -> raise Errors.Help
                | program_name :: l -> ignore program_name; l
        );
        arg

let main () =
        try
                begin
                        let arg = parse_arg () in

                        let verbose s = if arg.verbose then (print_string s; print_newline ()) else () in

                        (* Suit la liste de toutes les actions possibles. *)

                        let node_list =
                                if arg.read_lustre then
                                begin
                                        verbose "Reading Lustre input file.";
                                        Lustre.load_program arg.input arg.node_to_be_analysed
                                end
                                else []
                        in

                        if arg.compute_automata then
                        begin
                                Projection_apron.init ();
                                Formula_yices.full_simplify := true
                        end;

                        List.iter
                        (fun node ->
                                let variables =
                                        {
                                                node = node;
                                                predicates = None;
                                                logic_transition = None;
                                                logic_initialisation = None;
                                                invariant = None;
                                        }
                                in
                                let size = arg.invariant_options.Invariants.size in
                                do_node arg variables;
                                arg.invariant_options.Invariants.size <- size; (* En effet, il est possible que la recherche d’invariant fasse augmenter cette valeur. *)
                                ()
                        )
                        node_list;

              end
        with e -> Errors.catch e

let _ = main ()

