(* node_action.ml *)
(* Contient la fonction qui travaille à partir des arguments sur un nœud
 * en particulier. *)

open Flux


type argument =
        {
                mutable input : in_channel;
                mutable output : out_channel;

                mutable verbose : bool;

                mutable read_lustre : bool;
                mutable write_lustre : bool;

                mutable node_to_be_analysed : string option;

                mutable compute_predicates : bool;
                mutable level_predicates : int;
                mutable print_predicates : bool;

                mutable simplify_trivial_equation : bool;
                mutable simplify_if_blocs : bool;
                mutable expand_pre_blocs : bool;

                mutable compute_logic_node : bool;
                mutable print_logic_node : bool;

                mutable find_invariant : bool;
                mutable invariant_options : Invariants.arg;
                mutable print_invariant : bool;

                mutable compute_automata : bool;
                mutable print_automata : bool;
        }


type variables_node =
        {
                mutable node : Syntax.node;
                mutable predicates : Syntax.expression list option;
                mutable logic_transition : Syntax.expression option;
                mutable logic_initialisation : Syntax.expression option;
                mutable invariant : Syntax.expression list list option;
        }

let do_node arg variables =

        let verbose s = if arg.verbose then (print_string s; print_newline ()) else () in

        verbose ("Analysing node “" ^ Alpha.get_name variables.node.Syntax.id ^ "”.");

        if arg.expand_pre_blocs then
                begin
                        verbose "Expand pre blocs.";
                        Simplify.all_pre_to_one variables.node
                end;

        if arg.simplify_if_blocs then
                begin
                        verbose "Simplify if and arrow blocs.";
                        Simplify.extract_if_arrow_blocs variables.node
                end;

        if arg.simplify_trivial_equation then
                begin
                        verbose "Simplify trivial equations.";
                        variables.node <- Simplify.remove_trivial_equation variables.node
                end;

        if arg.write_lustre then
                begin
                        verbose "Printing Lustre output.";
                        print_newline ();
                        Printer.print_node variables.node
                end;

        if arg.compute_predicates then
                begin
                        Logic_node.set_level_research arg.level_predicates;
                        verbose "Computing predicates.";
                        variables.predicates <- Some (Logic_node.get_predicates variables.node)
                end;

        if arg.print_predicates then
                begin
                        verbose "Printing predicates.";
                        match variables.predicates with
                        | None -> ()
                        | Some l ->
                            List.iter (fun e ->
                                    print_string "predicate ( ";
                                    Printer.print_expression e;
                                    print_string " ) ;\n") l;
                                    print_newline ();
                end;

        if arg.compute_logic_node then
                begin
                        verbose "Computing the logical equivalent of a node.";
                        Logic_node.init_logic variables.node;
                        variables.logic_transition <- Some (Logic_node.logic_transition_of_node variables.node);
                        variables.logic_initialisation <- Some (Logic_node.logic_initialisation_of_node variables.node);
                end;

        if arg.print_logic_node then
                 begin
                         verbose "Printing the logical equivalent of a node.";
                         (match variables.logic_transition with
                         | None -> ()
                         | Some e ->
                             print_string "transition ( ";
                             Printer.print_expression e;
                             print_string " ) ;\n");
                         print_newline ();
                         (match variables.logic_initialisation with
                         | None -> ()
                         | Some e ->
                             print_string "initialisation ( ";
                             Printer.print_expression e;
                             print_string " ) ;\n");
                         print_newline ()
                 end;

        if arg.find_invariant then
                begin
                        verbose "Searching an invariant property for a node.";
                        match variables.predicates, variables.logic_transition, variables.logic_initialisation with
                        | Some [], _, _ -> failwith "no predicates founded."
                        | Some pred, Some transition, Some initialisation ->
                               variables.invariant <- Some (Solve.search pred transition initialisation arg.invariant_options)
                        | _, _, _ -> failwith "missings preliminary computations to compute an invariant."
                end;

        if arg.compute_automata then
                begin
                         verbose "Computing an equivalent automata for this node.";
                         let automata = Invariants.get_automata () in
                         Automata.set_automata variables.node automata;
                         Transitions.compute_conditions automata;
                         Automata.give_automata variables.node.Syntax.id automata
                end;

        if arg.print_invariant then
                 begin
                         verbose "Printing the found invariant.";
                         (match variables.invariant with
                         | None -> ()
                         | Some l ->
                               print_string "invariant : \n";
                               Printer.print_T l;
                               print_newline (); print_newline ())
                 end;

        if arg.print_automata then
                 begin
                         verbose "Printing automata.";
                         Printer.print_automata (Invariants.get_automata ())
                 end;

        verbose ""

