(* printer.ml *)
(* Contient toutes les fonctions de sortie *)

open Flux
open Syntax


let string_of_type = function
        | TBool -> "bool"
        | TInt -> "int"
        | TReal -> "real"
        | TTuple _ -> failwith "a tuple type should never be printed."

let print_type = function
        | TBool -> print_string "bool"
        | TInt -> print_string "int"
        | TReal -> print_string "real"
        | TTuple _ -> failwith "a tuple type should never be printed."

let rec print_expression = function
        | True, _, _ -> print_string "true"
        | False, _, _ -> print_string "false"
        | Int i, _, _ -> print_string (Mpzf.to_string i)
        | Real f, _, _ -> print_float (Mpqf.to_float f)
        | Ident i, _, _ -> print_string (Alpha.get_name i)

        | Not, Op [e], _ ->
              print_string "not "; print_expression e; print_string ""
        | Uminus, Op [e], _ ->
              print_string "-"; print_expression e; print_string ""
        | Int2Real, Op [e], _ ->
              print_string "int("; print_expression e; print_string ")"
        | Real2Int, Op [e], _ ->
              print_string "real("; print_expression e; print_string ")"

        | Arrow, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " -> ";
              print_expression e2;
              print_string ")"
        | And, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " and ";
              print_expression e2;
              print_string ")"
        | Or, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " or ";
              print_expression e2;
              print_string ")"
        | Xor, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " xor ";
              print_expression e2;
              print_string ")"
        | Impl, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " => ";
              print_expression e2;
              print_string ")"
        | Eq, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " = ";
              print_expression e2;
              print_string ")"
        | Neq, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " <> ";
              print_expression e2;
              print_string ")"
        | Gte, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " >= ";
              print_expression e2;
              print_string ")"
        | Gt, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " > ";
              print_expression e2;
              print_string ")"
        | Lte, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " <= ";
              print_expression e2;
              print_string ")"
        | Lt, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " < ";
              print_expression e2;
              print_string ")"
        | Div, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " / ";
              print_expression e2;
              print_string ")"
        | Mod, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " mod ";
              print_expression e2;
              print_string ")"
        | Minus, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " - ";
              print_expression e2;
              print_string ")"
        | Plus, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " + ";
              print_expression e2;
              print_string ")"
        | Times, Op [e1; e2], _ ->
              print_string "(";
              print_expression e1;
              print_string " * ";
              print_expression e2;
              print_string ")"

        | If, Op [e1; e2; e3], _ ->
              print_string "if ";
              print_expression e1;
              print_string " then ";
              print_expression e2;
              print_string " else ";
              print_expression e3

        | Nor, Op l, _ ->
              print_string "nor( ";
              let rec aux = function
                      | [] -> ()
                      | e :: l ->
                          print_expression e;
                          print_char ' ';
                          aux l
              in aux l;
              print_string ")"
        | Diese, Op l, _ ->
              print_string "#( ";
              let rec aux = function
                      | [] -> ()
                      | [e] -> print_expression e; print_char ' '
                      | e :: l ->
                          print_expression e;
                          print_string ", ";
                          aux l
              in aux l;
              print_string ")"
        | Tuple, Op l, _ ->
              print_string "(";
              let rec aux = function
                      | [] -> ()
                      | e :: l ->
                          print_expression e;
                          if l <> [] then print_string ", ";
                          aux l
              in aux l;
              print_string ")"

        | Call i, Op l, _ ->
              print_string (Alpha.get_name i);
              print_string "(";
              let rec aux = function
                      | [] -> ()
                      | e :: l ->
                          print_expression e;
                          if l <> [] then print_string ", ";
                          aux l
              in aux l;
              print_string ")"

        | _ -> failwith "problem of aritee while printing."


let print_node node =
        print_string ("node " ^ Alpha.get_name node.id ^ " (");
        let rec print_arg s = function
                | [] -> ()
                | i :: l ->
                     print_string (Alpha.get_name i ^ ": ");
                     print_type (Alpha.get_type i);
                     if l != [] then print_string (";" ^ s);
                     print_arg s l
        in print_arg " " node.input;
        print_string ") returns (";
        print_arg " " node.output;
        print_string ") ;\n";
        if node.temporary <> [] || node.memory <> [] then
        begin
                print_string "var\n";
                if node.temporary <> [] then (print_char '\t'; print_arg "\n\t" node.temporary; print_string " ;\n");
                if node.memory <> [] then (print_string
                (String.concat "" (List.map (fun (i, _, _) ->
                        "\t" ^ Alpha.get_name i ^ " : " ^ string_of_type (Alpha.get_type i) ^ " ;\n") node.memory)));
        end;
        print_string "let\n";
        print_string ("\tassert (" ^ Alpha.get_name node.assume ^ ") ;\n\n");
        let print_equation (l, e) =
                print_char '\t';
                let rec aux = function
                        | [] -> ()
                        | i :: l ->
                             print_string (Alpha.get_name i);
                             if l <> [] then print_string ", ";
                             aux l
                in aux l;
                print_string " = ";
                print_expression e;
                print_string " ;\n"
        in
        List.iter print_equation node.equations;
        if node.memory <> [] then print_newline ();
        List.iter (fun (id, id_suivi, decalage) ->
                print_string ("\t" ^ (Alpha.get_name id) ^ " = ");
                for i = 1 to decalage do print_string "pre "; done;
                print_string (Alpha.get_name id_suivi);
                print_string " ;\n"
        ) node.memory;
        print_string "tel.\n"

let rec print_T =
        let rec print_and = function
                | [] -> ()
                | [a] -> print_char '\t'; print_expression a; print_newline ()
                | a :: l -> print_char '\t'; print_expression a; print_string "\n\tand"; print_and l
        in function
        | [] -> print_string "\t\tfalse"
        | [[]] -> print_string "\t\ttrue"
        | [l] -> print_char '\t'; print_and l
        | l :: suite -> print_char '\t'; print_and l; print_string "or"; print_T suite

let print_automata automate =
        print_string "inputs:\n";
        List.iter (fun v -> print_string ("\t" ^ Alpha.get_name v ^ " : "); print_type (Alpha.get_type v); print_newline ()) automate.Automata.input;
        print_newline ();
        print_string "outputs:\n";
        List.iter (fun v -> print_string ("\t" ^ Alpha.get_name v ^ " : "); print_type (Alpha.get_type v); print_newline ()) automate.Automata.output;
        print_newline ();
        print_string "states:\n";
        let print_state i state =
                print_string ("\tstate " ^ string_of_int i ^ ":\n");
                let first = ref true in
                let f e =
                        print_string "\t\t";
                        if !first then first := false else print_string "and";
                        print_char '\t';
                        print_expression e;
                        print_newline ()
                in
                List.iter f state;
                if state = [] then print_string "\t\ttrue\n";
                print_string "\tThat yields:\n\t\t";
                print_expression automate.Automata.output_prop.(i);
                print_newline ()
        in
        Array.iteri print_state automate.Automata.states;
        print_newline ();
        print_string "initial states:\n";
        List.iter (fun i -> print_string ("\t— state " ^ string_of_int i ^ ".\n")) automate.Automata.init_states;
        print_newline ();
        print_string "transitions:\n";
        List.iter (fun (s, e, cond_in) -> print_string ("\t— state " ^ string_of_int s ^ " to state " ^ string_of_int e ^ " when "); print_expression cond_in; print_string ".\n") automate.Automata.transitions;
        ()

