(* logic_node.ml *)
(* Contient des fonctions de conversions d’un node vers une formule logique. *)

open Syntax


let max_level = 4
(* Niveau 0 : il ne renvoit pas de prédicats. *)
(* Niveau 1 : il ne renvoit que les prédicats présents dans les tests des blocs if. *)
(* Niveau 2 : il renvoit en plus tous les booléens et leurs négations. *)
(* Niveau 3 : il renvoit en plus toutes les égalités suggérées par les équations. *)
(* Niveau 4 : il prend en plus des égalités des if les inégalités correspondantes. *)

let level = ref max_level

let set_level_research i =
        level := i


let prime_ident = ref [] (* Va contenir des couples (id, id') (J’aurais pu
                          * utiliser une nouvelle Hashtbl, mais je ne voulais
                          * pas trop compliquer les choses comme ça.) *)

let states_call = ref [] (* Va contenir les variables ajoutées pour les appels
de fonction. *)

let prime_of id = try List.assoc id !prime_ident
        with Not_found -> failwith ("error while searching the prime identifier of “" ^ Alpha.get_name id ^ "”.")

let create_prime_variable l =
        let f id =
                let new_id = Alpha.get_new_id (Alpha.get_new_name (Alpha.get_name id ^ "_prime")) (Alpha.get_type id) in
                prime_ident := (id, new_id) :: !prime_ident
        in List.iter f l


let rec primer = function
        | Ident id, _, t -> Ident (prime_of id), Op [], t
        | call, Op l, t -> call, Op (List.map primer l), t


let precompute_call node =
        let search_for_call = function
                | (left, (Call f, Op arg, ty)) ->
                       let name = Alpha.get_new_name (Alpha.get_name f ^ "_state_to_obtain_" ^ String.concat "_" (List.map Alpha.get_name left)) in
                       let i = Alpha.get_new_id name TInt in
                       let pre_i = Alpha.get_new_id ("pre_" ^ name) TInt in
                       let make_id i = Ident i, Op [], TInt in
                       node.temporary <- i :: node.temporary;
                       node.memory <- (pre_i, i, 1) :: node.memory;
                       states_call := i :: pre_i :: !states_call;
                       (left, (Call f, Op (make_id i :: make_id pre_i :: arg), ty))
                | a -> a
        in
        node.equations <- List.map search_for_call node.equations;
        ()


let egal prime left expr =
        let p = if prime then prime_of else (fun x -> x) in
        let primer = if prime then primer else (fun x -> x) in
        match left with
        | [a] -> (Eq, Op [Alpha.var_ident (p a) ; primer expr], TBool)
        | l -> match expr with
                | Tuple, Op le, _ ->
                      let rec aux = function
                              | [], [] -> from_bool true
                              | [a], [e] -> (Eq, Op [Alpha.var_ident (p a) ; primer e], TBool)
                              | a :: l, e :: le ->
                                              (And, Op [(Eq, Op [Alpha.var_ident (p a) ; primer e], TBool) ; aux (l, le)], TBool)
                              | _ -> failwith "aritee failed when converting a node to a logic transition expression."
                      in aux (l, le)
                | (If | Arrow | Call _), _, _ -> failwith "there stay Call, If or Arrow blocs in a node while trying to convert it to a logic transition expression."
                | _ -> failwith "aritee failed when converting a node to a logic transition expression."


let logic_transition_of_equation = function
        | left, (If, Op [e1 ; e2 ; e3], _) ->
                        make_or
                        [
                                make_and [primer e1 ; (egal true left e2)] ;
                                make_and [(Not, Op [primer e1], TBool) ; (egal true left e3)]
                        ]
        | left, (Arrow, Op [_ ; e2], _) -> egal true left e2
        | left, (Call f, Op ((Ident state, _, _) :: (Ident pre_state, _, _) :: args), _) ->
                        Automata.transform_expression state pre_state (Automata.get_automata f) args left
        | left, expr -> egal true left expr

let logic_initialisation_of_equation = function
        | left, (If, Op [e1 ; e2 ; e3], _) ->
                        make_or
                        [
                                make_and [e1 ; (egal false left e2)] ;
                                make_and [(Not, Op [e1], TBool) ; (egal false left e3)]
                        ]
        | left, (Arrow, Op [e1 ; _], _) -> egal false left e1
        | left, (Call f, Op ((Ident state, _, _) :: (Ident pre_state, _, _) :: args), _) ->
                        Automata.initialisation_automate state (Automata.get_automata f)
        | left, expr -> egal false left expr


let init_logic node =
        precompute_call node;
        create_prime_variable node.input;
        create_prime_variable node.output;
        create_prime_variable node.temporary;
        create_prime_variable !states_call;
        create_prime_variable (List.map (fun (id, _, _) -> id) node.memory);
        ()

let logic_transition_of_node node =
        let rec memory_transition = function
                | [] -> from_bool true
                | (id, id_suiv, 1) :: [] -> (Eq, Op [Alpha.var_ident (prime_of id) ; Alpha.var_ident id_suiv], TBool)
                | (id, id_suiv, 1) :: l -> (And, Op [(Eq, Op [Alpha.var_ident (prime_of id) ; Alpha.var_ident id_suiv], TBool) ; memory_transition l], TBool)
                | _ -> failwith "presence of shift greater than one while calling logic_transition_of_node."
        in
        let rec make_and = function
                | [] -> from_bool true
                | a :: [] -> a
                | a :: l -> (And, Op [a ; make_and l], TBool)
        in
        let equ_logic = List.map logic_transition_of_equation node.equations in
        make_and [Alpha.var_ident node.assume ; make_and equ_logic ; memory_transition node.memory]

let logic_initialisation_of_node node =
        let rec aux = function
                | [] -> from_bool true
                | a :: [] -> a
                | a :: l -> match aux l with (* Juste pour éviter trop de « and true ». *)
                            | True, _, _ -> a
                            | aux_l ->
                                    match a with
                                    | True, _, _ -> aux_l
                                    | _ -> (And, Op [a ; aux_l], TBool)
        in
        aux (Alpha.var_ident node.assume :: List.map logic_initialisation_of_equation node.equations)


let get_predicates node =
        (* Renvoit tous les prédicats à possiblement prendre en compte du nœud. *)
        let rec search_predicate (call, Op l, _) =
                (List.concat (List.map search_predicate l))
                @ match call with
                | If ->
                                begin
                                        match l with
                                        | [(Eq | Lte | Lt | Gte | Gt), Op [((_, _, (TInt | TReal)) as e1); e2], TBool; _; _] when !level >= 4 -> List.map (fun s -> (s, Op [e1 ; e2], TBool)) [Eq ; Lte ; Lt ; Gte ; Gt]
                                        | ([(Not, Op [e1], TBool) ; _ ; _] | [e1; _ ; _ ]) -> [e1 ; (Not, Op [e1], TBool)]
                                        | _ -> failwith "internal problem of aritee while searching predicates in a node"
                                end
                | Nor -> l
                | Diese -> l
                | _ -> []
        in
        let rec search_equation_predicates left = function
                | Tuple, Op l, _ ->
                                let listify = List.map (fun a -> [a]) in
                                List.concat (List.map2 search_equation_predicates (listify left) l)
                | ((Ident _ | Real _ | Int _), _, _) as a ->
                                (match left with
                                | [b] -> [Eq, Op [a ; Alpha.var_ident b], TBool]
                                | _ -> failwith "problem of aritee while searching predicates.")
                | ((Uminus | Int2Real | Real2Int | Div | Mod | Minus | Plus | Times), _, (TInt | TReal)) as a ->
                                (match left with
                                | [b] when !level >= 4 -> List.map (fun s -> (s, Op [a ; Alpha.var_ident b], TBool)) [Eq ; Gte ; Gt ; Lte ; Lt]
                                | [b] -> [Eq, Op [a ; Alpha.var_ident b], TBool]
                                | _ -> failwith "problem of aritee while searching predicates.")
                | If, Op [_; e1; e2], _ -> search_equation_predicates left e1 @ search_equation_predicates left e2
                | Call _, _, _ -> []
                | _, Op l, _ -> List.concat (List.map (search_equation_predicates left) l)
        in
        let rec equation_predicates = function
                | [] -> []
                | (left, right) :: l ->
                                search_equation_predicates left right @ equation_predicates l
        in
        let rec search_boolean node =
                let aux l =
                        List.fold_left (fun l a -> if Alpha.get_type a = TBool then let v = Alpha.var_ident a in v :: (Not, Op [v], TBool) :: l else l) [] l
                in
                aux node.input
                @ aux node.output
                @ aux node.temporary
                @ aux (List.map (fun (a, _, _) -> a) node.memory)
        in
        let simplify l =
                let rec uniq l = function
                        | [] -> l
                        | a :: b :: s when a = b -> uniq l (b :: s)
                        | a :: s -> uniq (a :: l) s
                in
                uniq [] (List.sort compare l)
        in
        let level n l = if !level >= n then l else [] in
        simplify
        (
                level 1 (List.concat (List.map (fun (_, e) -> search_predicate e) node.equations))
                @ level 3 (equation_predicates node.equations)
                @ level 2 (search_boolean node)
        )

