(* simplify.ml *)
(* Contient des fonctions simplifiant les nœuds *)

open Syntax

(* Ci dessous quelques lignes pour une union find, structure pratique pour la suite. *)
let union_find = Hashtbl.create 10

let rec find_union x =
        match Hashtbl.find union_find x with
        | y when x = y -> x
        | y -> find_union y

let create_union x =
        try let _ = Hashtbl.find union_find x in () with
        | Not_found -> Hashtbl.add union_find x x

let merge_union x y =
        Hashtbl.add union_find (find_union x) (find_union y)

(* fin de l’union find. *)

let rec replace_all = function
        | Ident x, Op [], ty -> Ident (find_union x), Op [], ty
        | call, Op l, ty -> call, Op (List.map replace_all l), ty

let rec replace_all_memory = function
        | [] -> []
        | (id, id_suiv, dec) :: l -> (find_union id, find_union id_suiv, dec) :: replace_all_memory l

let init node =
        Hashtbl.clear union_find;
        List.iter (fun (x, _, _) -> create_union x) node.memory;
        List.iter create_union node.input;
        List.iter create_union node.output;
        List.iter create_union node.temporary

let actualize node =
        let rec concat_option = function
                | [] -> []
                | None :: l -> concat_option l
                | Some x :: l -> x :: concat_option l
        in
        let remove x =
                Alpha.remove x;
                node.temporary <- List.filter ((<>) x) node.temporary
        in
        let rec not_in x = function
                | y :: _ when y = x -> false
                | _ :: l -> not_in x l
                | [] -> true
        in
        node.equations <- concat_option (List.map
        (function
                | [x], (Ident _, Op [], _) when find_union x <> x && not_in x node.output -> remove x; None
                | ([x], (Ident _, Op [], _)) as equ when find_union x <> x -> Some equ (* Car x est dans la liste des variables de sortie. *)
                | (l, expr) -> Some (List.map find_union l, replace_all expr)
        ) node.equations);
        node.memory <- replace_all_memory node.memory;
        node.assume <- find_union node.assume;
        ()

let remove_trivial_equation node =
        init node;
        let unfold_equation = function
                | l, (Tuple, Op v, _) -> List.map2 (fun var -> fun v -> ([var], v)) l v
                | a -> [a]
        in
        let check_equation = function
                | [x], (Ident y, Op [], _) -> merge_union x y
                | l, _ -> ()
                (* Si x n’est pas trivial, disons que c’est son propre père. *)
        in
        node.equations <- List.concat (List.map unfold_equation node.equations);
        List.iter check_equation node.equations;
        actualize node;
        node


let extract_if_arrow_blocs node =
        let new_variables = ref [] in
        let new_equations = ref [] in
        let rec travel_expression lx = function (* lx est la liste de noms de variables
        dont on parcours actuellement l’équation. *)
                | If, Op [e1 ; e2 ; e3], ty ->
                      let name_if = List.map (fun x -> Alpha.get_new_name (x ^ "_if_part")) lx in
                      let if_part = match ty with
                                    | TTuple lty -> List.map2 (fun t -> fun name -> Alpha.get_new_id name t) lty name_if
                                    | t -> List.map (fun name -> Alpha.get_new_id name t) name_if
                      in
                      new_variables := if_part @ !new_variables;
                      let e1' = travel_expression (List.map (fun name -> name ^ "_test") name_if) e1 in
                      let e2' = travel_expression (List.map (fun name -> name ^ "_true") name_if) e2 in
                      let e3' = travel_expression (List.map (fun name -> name ^ "_false") name_if) e3 in
                      new_equations := (if_part, (If, Op [e1' ; e2' ; e3'], ty)) :: !new_equations;
                      (match List.map (fun part -> Alpha.var_ident part) if_part with
                      | [a] -> a
                      | l -> Tuple, Op l, ty)
                | Arrow, Op [e1 ; e2], ty ->
                      let name_arrow = List.map (fun x -> Alpha.get_new_name (x ^ "_arrow_part")) lx in
                      let arrow_part = match ty with
                                       | TTuple lty -> List.map2 (fun t -> fun name -> Alpha.get_new_id name t) lty name_arrow
                                       | t -> List.map (fun name -> Alpha.get_new_id name ty) name_arrow in
                      new_variables := arrow_part @ !new_variables;
                      let e1' = travel_expression (List.map (fun name -> name ^ "_start") name_arrow) e1 in
                      let e2' = travel_expression (List.map (fun name -> name ^ "_following") name_arrow) e2 in
                      new_equations := (arrow_part, (Arrow, Op [e1' ; e2'], ty)) :: !new_equations;
                      (match List.map (fun part -> Alpha.var_ident part) arrow_part with
                      | [a] -> a
                      | l -> Tuple, Op l, ty)
                | Call f, Op args, ty ->
                      let name_function = List.map (fun x -> Alpha.get_new_name (x ^ "_fun_" ^ Alpha.get_name f ^ "_part")) lx in
                      let function_part =
                              List.map2 (fun t name -> Alpha.get_new_id name t) (List.map Alpha.get_type ((Automata.get_automata f).Automata.output)) name_function
                      in
                      new_variables := function_part @ !new_variables;
                      let arg' = List.map (travel_expression name_function) args in
                      new_equations := (function_part, (Call f, Op arg', ty)) :: !new_equations;
                      (match List.map (fun part -> Alpha.var_ident part) function_part with
                      | [a] -> a
                      | l -> Tuple, Op l, ty)
                | call, Op l, ty -> call, Op (List.map (travel_expression lx) l), ty
        in
        let rec travel_equations = function
                | l, expr ->
                       l, travel_expression (List.map Alpha.get_name l) expr
        in
        node.equations <- (List.map travel_equations node.equations);
        node.equations <- node.equations @ !new_equations;
        node.temporary <- node.temporary @ !new_variables


let all_pre_to_one node =
        init node;
        let table = Hashtbl.create 10 in
        let rec check_memory = function
                | [] -> ()
                | (id, id_suiv, dec) :: l ->
                                try let other = Hashtbl.find table (id_suiv, dec) in merge_union id other with
                                | Not_found ->
                                        let rec aux = function
                                                | 0 -> id_suiv
                                                | n ->
                                                        try Hashtbl.find table (id_suiv, n) with
                                                        | Not_found ->
                                                        let name = Alpha.get_new_name ("pre_" ^ Alpha.get_name (aux (n - 1))) in
                                                        let new_id = Alpha.get_new_id name (Alpha.get_type id_suiv) in
                                                        Hashtbl.add table (id_suiv, n) new_id; create_union new_id;
                                                        new_id
                                        in let new_id = aux dec in merge_union new_id id;
                                        check_memory l
        in
        check_memory node.memory;
        let create_equation id id_suiv = function
                | 1 -> (id, id_suiv, 1)
                | n -> let id_suiv1 = Hashtbl.find table (id_suiv, n - 1) in (id, id_suiv1, 1)
        in
        node.memory <- (
                let ret = ref [] in
                Hashtbl.iter (fun (id_suiv, n) id -> ret := create_equation id id_suiv n :: !ret) table;
                !ret);
        actualize node

