(* invariants.ml *)
(* Contient des fonctions calculant les modes de fonctionnement d’un nœud. *)

open Smt
open Yices
open Syntax
open Context
open Utils
open Automata

type arg =
        {
                mutable just_a_little_verbose : bool;
                mutable verbose : bool;
                mutable print : string -> unit;
                mutable print_T : Syntax.expression list list -> unit;

                mutable minimal : bool;
                mutable size : int;

                mutable search_better_size : bool;
                mutable max_size : int;

                mutable removal_of_permutation : bool;
                mutable removal_of_subsumed_disjuncts : bool;
                mutable satisfiable : bool;
                mutable variante_eviter_seg_fault_yices : bool;

                mutable find_better_strictly : bool;

                mutable expand_invariant : bool;

                mutable generate_automata : bool;
        }

let default =
        {
                just_a_little_verbose = false;
                verbose = false;
                print = (fun _ -> ());
                print_T = (fun _ -> ());

                minimal = true;
                size = 1;

                search_better_size = false;
                max_size = max_int;

                removal_of_permutation = true;
                removal_of_subsumed_disjuncts = false;
                satisfiable = false;
                variante_eviter_seg_fault_yices = false;

                find_better_strictly = true;

                expand_invariant = false;

                generate_automata = false;
        }


let context_H = mk_context ()
let id_H = get_new_id_context ()

let context_F = mk_context ()
let id_F = get_new_id_context ()

let b_H = ref (Array.make_matrix 0 0 yices_expr)
let b_F = ref (Array.make_matrix 0 0 yices_expr)

let tm, tn = ref 0, ref 0

let pi = ref (Array.make 0 (from_bool true))
let pi_context_F = ref (Array.make 0 yices_expr) (* Contient les pi_j *)
let pi_context_F_prime = ref (Array.make 0 yices_expr) (* Contient les pi_j, dont les variables ont été primées. *)

let not_F = ref yices_expr
let added_constraints_on_H = ref [] (* bool list list list ref *)

let trans_form = ref yices_expr
let init_form = ref yices_expr

let automata = Automata.empty_automata ()


let init n m =
        push context_F; (* Ces push doivent être popés par reset. *)
        push context_H;
        tm := m;
        tn := n;
        b_H := Array.make_matrix n m yices_expr;
        b_F := Array.make_matrix n m yices_expr;
        for i = 0 to n - 1 do
                for j = 0 to m - 1 do
                       let b_i_j_H = mk_fresh_bool_var context_H in !b_H.(i).(j) <- b_i_j_H;
                       let b_i_j_F = mk_fresh_bool_var context_F in !b_F.(i).(j) <- b_i_j_F;
                       ()
               done;
        done;
        pi := Array.make m (from_bool true);
        pi_context_F := Array.make m yices_expr;
        pi_context_F_prime := Array.make m yices_expr;
        Automata.cp_automata automata (Automata.empty_automata ());
        automata.states <- Array.make n [];
        automata.output_prop <- Array.make n (from_bool true);
        added_constraints_on_H := [];
        ()


let set_transition_formula form =
        automata.transition_formula <- form;
        trans_form := convert_expr context_F id_F form;
        ()

let set_initial_formula form =
        init_form := convert_expr context_F id_F form;
        ()

let init_formula () =
        (* On va créer dans le contexte F la condition not F(b, sigma). (en laissant libre les variables b_ij et sigma). *)
        let t = create_T context_F !tn !tm !b_F !pi_context_F in
        let t_prime = create_T context_F !tn !tm !b_F !pi_context_F_prime in
        not_F := (mk_not context_F (implique context_F !trans_form
                (mk_and_nsf context_F [
                implique context_F !init_form t;
                implique context_F t
                 (t_prime)
        ])))

let rec init_context_from_predicates n j = function
        | [] -> ()
        | pi_i :: l ->
               !pi.(!tm - j) <- pi_i;
               !pi_context_F.(!tm - j) <- convert_expr context_F id_F pi_i;
               !pi_context_F_prime.(!tm - j) <- convert_expr context_F id_F (Logic_node.primer pi_i);
               init_context_from_predicates n (j - 1) l

let init_constraint arg n m = (* Ajoute des contraintes dans le but d’optimiser le programme final. *)
        let o a b = mk_or_nsf context_H [a ; b] in
        let an a b = mk_and_nsf context_H [a ; b] in
        let no a = mk_not context_H a in
        let b i j = !b_H.(i).(j) in
        (* Removal of Permutations *) 
        if arg.removal_of_permutation then
        begin
                if log (float_of_int n) /. log 2. >= float_of_int m then failwith "there is not enough predicates to found a solution with so many clauses.";
                let rec l i j =
                        match i, j with
                        | _, v when v = m -> mk_false context_H
                        | i, j -> o (an (no (b i j)) (b (i + 1) j)) (an (o (no (b i j)) (b (i + 1) j)) (l i (j + 1))) (* C’est la formule 3.2 du papier. *)
                in
                for i = 0 to n - 2 do
                        assert_expr context_H (l i 0)
                done
        end;
        (* Removal of Subsumed Disjuncts *)
        if arg.removal_of_subsumed_disjuncts then
        begin
                if m = 0 then failwith "there is no predicates : the algorithm cannot works.";
                let id_context = ref 0 in
                let c i =
                    let conv e =
                        convert_expr context_H !id_context e
                    in
                    create_c context_H m !b_H.(i) (Array.map conv !pi)
                in
                let all_not_expect i0 = (* /\_{i <> i0}, not C_i *)
                        let rec aux = function
                                | i when i = i0 -> aux (i + 1)
                                | i when i = n -> []
                                | i -> (mk_not context_H (c i)) :: aux (i + 1)
                        in aux 0
                in
                let subsumed_disjuncts i =
                        mk_and_nsf context_H (c i :: all_not_expect i)
                in
                for i = 0 to n - 1 do
                        id_context := get_new_id_context ();
                        assert_expr context_H (subsumed_disjuncts i)
                done
        end
        else (* Removal of Subsumed Disjuncts l’implique. *)
        if arg.satisfiable then
        begin
                let f_satisfiable id_cont i =
                        let conv = convert_expr context_H id_cont
                        in
                        create_c context_H m !b_H.(i) (Array.map conv !pi)
                in
                for i = 0 to n - 1 do
                        assert_expr context_H (f_satisfiable (get_new_id_context ()) i)
                done
        end;
        ()

let add_constraint_to_found_a_better_solution_than e =
        (* On veut en plus ajouter en contrainte de H qu’il existe au moins une solution de e qui ne satisfasse pas la nouvelle contrainte (sinon, ça ne sert à rien d’en trouver une plus compliqué.) *)
        let id_context = get_new_id_context () in (* Ceci permet de déclarer facilement de nouvelles variables. *)
        let conv e =
                convert_expr context_H id_context e
        in
        let nouvelle =
                create_T context_H !tn !tm !b_H (Array.map conv !pi)
        in
        let ancienne =
                mk_or_nsf context_H (List.map (fun l -> mk_and_nsf context_H (List.map conv l)) e)
        in
        let new_constraint =
                mk_and_nsf context_H [ancienne; mk_not context_H nouvelle]
        in
        assert_expr context_H new_constraint; (* FIXME: Seg fault à cette ligne là sur quelques exemples ! Des variables non déclarées peut-être ? *)
        ()

let test_better sol1 sol2 =
        let c = mk_context () in
        let id_c = get_new_id_context () in

        let conv e = convert_expr c id_c e in
        let convert sol =
                mk_or_nsf c (List.map (fun l -> mk_and_nsf c (List.map conv l)) sol)
        in
        let e1 = convert sol1 in
        let e2 = convert sol2 in

        assert_expr c (mk_and_nsf c [e2; mk_not c e1]);
        match check c with
        | L_UNKNOWN -> raise Errors.Yices_unknown
        | L_TRUE -> true
        | L_FALSE -> false

let convert_to_matrix model =
        let b = Array.make_matrix !tn !tm false in
        for i = 0 to !tn - 1 do
                for j = 0 to !tm - 1 do
                        switch_value_in model !b_H.(i).(j)
                        (fun () -> b.(i).(j) <- true)
                        (fun () -> b.(i).(j) <- false)
                done
        done;
        b

let recreate_expression_from_b_values b = (* C’est cette fonction qui créé le résultat final. *)
        let c i =
                remove_option
                (
                        let f j =
                                if b.(i).(j - 1)
                                then Some !pi.(j - 1)
                                else None
                        in
                        create_list f !tm
                )
        in
        automata.states <- Array.mapi (fun i _ -> c i) automata.states;
        create_list (fun i -> automata.states.(i - 1)) !tn

let recreate_expression_from_model model =
        recreate_expression_from_b_values (convert_to_matrix model)

let assert_b_i_j model = (* Va donner dans le contexte F les valeurs des différents booléens du contexte H *)
        for i = 0 to !tn - 1
        do
                for j = 0 to !tm - 1
                do
                        switch_value_in model !b_H.(i).(j)
                        (fun () -> assert_expr context_F !b_F.(i).(j))
                        (fun () -> assert_expr context_F (mk_not context_F !b_F.(i).(j)))
                done
        done


let evaluate_constraint_on_H_in sigma constraints =
        if constraints = [] then mk_true context_H else
        let pi = Array.map (model_expr_to_context_bool context_H sigma) !pi_context_F in
        let j = ref 0 in
        let atome = function
                | false -> incr j; mk_true context_H
                | true -> incr j; pi.(!j - 1)
        in
        let create_part_cons cons = (* cons est de type bool list list *)
                mk_or_nsf context_H (List.map (fun l -> j := 0; mk_and_nsf context_H (List.map atome l)) cons)
        in
        implique context_H (mk_or_nsf context_H (create_list (fun i -> mk_and_nsf context_H (create_list (fun j -> implique context_H !b_H.(i - 1).(j - 1) pi.(j - 1)) !tm)) !tn))
        (mk_and_nsf context_H (List.map create_part_cons constraints))

let adding_constraints_on_H model =
        let eval i j =
                switch_value_in model !b_H.(i - 1).(j - 1)
                (fun () -> true)
                (fun () -> false)
        in
        create_list (fun i -> create_list (fun j -> eval i j) !tm) !tn

let create_assert_F_Sigma_sigma sigma = (* Va créer la contrainte F[Sigma/sigma] dans le contexte H. *)
         (* On a F = (Init(sigma) => T(sigma)) /\ (T(sigma) => T(sigma') /\ ... où le reste est l’ensemble des contraintes imposées par les invariants trouvés précédemment. *)
         (* Attention : sigma est un modèle du contexte F. *)
        let t_sigma = create_T context_H !tn !tm !b_H (Array.map (model_expr_to_context_bool context_H sigma) !pi_context_F) in
        let t_sigma_prime = create_T context_H !tn !tm !b_H (Array.map (model_expr_to_context_bool context_H sigma) !pi_context_F_prime) in
        mk_and_nsf context_H [
                implique context_H (model_expr_to_context_bool context_H sigma !init_form) t_sigma;
                implique context_H t_sigma t_sigma_prime;
                evaluate_constraint_on_H_in sigma !added_constraints_on_H
        ]

let constraint_find_better arg b_values = (* Va imposer de nouvelles contraintes sur B (i.e. le contexte H) pour trouver une solution un peu plus petite. *)
    if arg.find_better_strictly then
    begin (* Au moins une solution qui marchait dans l’ancien, mais pas dans le nouveau.*)
        let id_context = get_new_id_context () in
        let c b i = (* /\_j b_ij => pi_j *)
                let atome j = (* b_ij => pi_j *)
                        implique context_H (b i j) (convert_expr context_H id_context !pi.(j - 1))
                in
                mk_and_nsf context_H (create_list atome !tm)
        in
        assert_expr context_H
        (
            mk_and_nsf context_H
            (
                let t b =
                    mk_or_nsf context_H (create_list (c b) !tn)
                in
                [
                        t (fun i j -> context_bool context_H b_values.(i - 1).(j - 1));
                        mk_not context_H (t (fun i j -> !b_H.(i - 1).(j - 1)))
                ]
            )
        )
    end
    else
    begin (* Au moins un booléen plus grand. *)
        let better i j =
                if b_values.(i).(j)
                then mk_false context_H
                else !b_H.(i).(j)
        in
        let rec aux l i j =
                match i, j with
                | _, -1 -> l
                | -1, j -> aux l (!tn - 1) (j - 1)
                | i, j -> aux (better i j :: l) (i - 1) j
        in
        assert_expr context_H
        (
                mk_or_nsf context_H (aux [] (!tn - 1) (!tm - 1))
        )
    end

let create_new_constraint_to_be_added_to_not_F model =
        (* Va imposer la contrainte de trouver une solution plus petite au sens de l’inclusion : not (T[B/b] => T[B’/b]) = T[B/b] /\ not T[B’/b]. *)
        (* B’ est ce que l’on vient de trouver (extrait du modèle donc), et B est la nouvelle matrice. *)
        (* Attention, model est un modèle de H. *)
        let nc i =
                let l = (create_list (fun j -> implique context_F (model_expr_to_context_bool context_F model !b_H.(i - 1).(j - 1)) !pi_context_F.(j - 1)) !tm) in
                mk_and_nsf context_F l
        in
        let t_B_prime_b =
                mk_or_nsf context_F (create_list nc !tn)
        in
        mk_and_nsf context_F
        [
                create_T context_F !tn !tm !b_F !pi_context_F;
                mk_not context_F t_B_prime_b
        ]

let expand_invariant_found model = (* Va ajuster tous les booléens b_ij que l’on peut fixer à vrai à vrai sans changer le sens de l’invariant trouvé. *)
        (* Cette fonction retourne la matrice des booléens b_ij. *)
        let b = Array.make_matrix !tn !tm false in
        for i = 0 to !tn - 1 do
                push context_F; (* Ce push est popé après la boucle. *)
                for j = 0 to !tm - 1
                do (* On met dans le modèle tous les cas de la conjonction. *)
                        switch_value_in model !b_H.(i).(j)
                        (fun () -> assert_expr context_F !pi_context_F.(j))
                        (fun () -> ())
                done;
                for j = 0 to !tm - 1 do
                        switch_value_in model !b_H.(i).(j)
                        (fun () -> b.(i).(j) <- true) (* Il était déjà vrai : on le garde. *)
                        (fun () -> (* On regarde si on peut le prendre en plus (s’il est impliqué par les autres) *)
                                push context_F; (* Dans les deux cas, ce push sera popé. *)
                                assert_expr context_F (mk_not context_F !pi_context_F.(j));
                                match check context_F with
                                | L_UNKNOWN -> raise Errors.Yices_unknown
                                | L_TRUE -> (* Cas non intéressant : pi_j peut 6etre fausse. *)
                                                pop context_F
                                | L_FALSE -> (* Cas où on peut ajouter cette expression à la clause. *)
                                                pop context_F;
                                                b.(i).(j) <- true
                        )
                done;
                pop context_F
        done;
        b

let regenerate_automata b =
        let c = Array.make !tn yices_expr in
        let c_prime = Array.make !tn yices_expr in
        let mk_c i pi =
                match
                        remove_option
                        (
                                let f j =
                                        if b.(i).(j - 1)
                                        then Some pi.(j - 1)
                                        else None
                                in
                                create_list f !tm
                        )
                with
                | [] -> mk_true context_F
                | l -> mk_and_nsf context_F l
        in
        for i = 0 to !tn - 1 do
                c.(i) <- mk_c i !pi_context_F;
                c_prime.(i) <- mk_c i !pi_context_F_prime;
        done;
        let trans = ref [] in
        for i = 0 to !tn - 1 do
                for j = 0 to !tn - 1 do
                        push context_F; (* Ce push est popé juste après. *)
                        assert_expr context_F !trans_form;
                        assert_expr context_F c.(i);
                        assert_expr context_F c_prime.(j);
                        (match check context_F with
                        | L_UNKNOWN -> raise Errors.Yices_unknown
                        | L_FALSE -> ()
                        | L_TRUE -> (* Il est possible de passer de l’état i à l’état j. *)
                                        trans := (i, j, from_bool true) :: !trans
                        );
                        pop context_F
                done
        done;
        automata.transitions <- !trans;
        let inits = ref [] in
        for i = 0 to !tn - 1 do
                push context_F;
                assert_expr context_F !init_form;
                assert_expr context_F c.(i);
                (match check context_F with
                | L_UNKNOWN -> raise Errors.Yices_unknown
                | L_FALSE -> ()
                | L_TRUE -> inits := i :: !inits;
                );
                pop context_F
        done;
        automata.init_states <- !inits;
        ()

let print_model_H arg model =
        arg.print "Satisfying assignement for the matrix B found :\n\t";
        let first = ref true in
        for i = 0 to !tn - 1 do
                for j = 0 to !tm - 1 do
                        if not !first then arg.print ", "
                        else first := false;
                        let v = evaluate_in_model model !b_H.(i).(j) in
                        arg.print ("B[" ^ string_of_int i ^ ", " ^ string_of_int j ^ "] = " ^ string_of_lbool v)
                done;
                if i <> !tn - 1 then arg.print ",\n\t";
                first := true
        done;
        arg.print ".\n";
        arg.print "This correspond to the invariant-candidate :\n";
        arg.print_T (recreate_expression_from_model model);
        arg.print "\n"


let print_model_F arg model =
        arg.print "Satisfying assignement for the memory Sigma found :\n\t";
        let first = ref true in
        let print_val x = function
                | TBool -> string_of_lbool (get_value model x)
                | TInt -> Z.to_string (get_mpz_value model x)
                | TReal -> Q.to_string (get_mpq_value model x)
                | TTuple _ -> failwith "a variable shouldn’t have a tuple type."
        in
        let f id decl =
                if !first then (arg.print "("; first := false)
                else arg.print ", ";
                arg.print (Alpha.get_name id ^ " = " ^ print_val decl (Alpha.get_type id))
        in
        Alpha.iter_on_yices_var id_F f;
        arg.print ")\n"


let rec solve arg last_result =
        if arg.verbose then arg.print "Searching an invariant candidate.\n";
        match check context_H with
        | L_UNKNOWN -> raise Errors.Yices_unknown
        | L_FALSE ->
                 if arg.verbose then
                 begin
                         arg.print "The constraints are too high : no solution can exist within them.\n";
                         arg.print "Using the last invariant found.\n\n"
                 end;
                 last_result ()
        | L_TRUE ->
                 let model = get_model context_H in (* On pourra ensuite puiser dans ce modèle les valeurs des différents booléens. *)
                 if arg.verbose then print_model_H arg model;
                 push context_F; (* Dans les deux cas du match, ce push est popé. *)
                 assert_b_i_j model;
                 assert_expr context_F !not_F;
                 if arg.verbose then arg.print "Asking Yices to know if it is an inductive invariant.\n";
                 match check context_F with (* On va tester le \not F[B/b]. *)
                 | L_UNKNOWN -> raise Errors.Yices_unknown
                 | L_FALSE -> (* On a une solution : la solution B ! *)
                                 (* On regarde si on peut trouver mieux. *)
                                 pop context_F;
                                 if arg.verbose then arg.print "This assignment for B is an inductive invariant.\n";
                                 let b_values =
                                         if arg.expand_invariant then
                                         begin
                                                if arg.verbose then arg.print "Expanding invariant to:\n";
                                                let m = expand_invariant_found model in
                                                if arg.verbose then arg.print_T (recreate_expression_from_b_values m);
                                                m
                                         end
                                         else convert_to_matrix model
                                 in
                                 if arg.minimal then
                                 begin
                                         if arg.verbose then arg.print "Trying to find a smaller invariant.\n\n";
                                         let new_constraint = create_new_constraint_to_be_added_to_not_F model in
                                         not_F := mk_or_nsf context_F [!not_F; new_constraint];
                                         added_constraints_on_H := (adding_constraints_on_H model) :: !added_constraints_on_H;
                                         constraint_find_better arg b_values;
                                         solve arg (fun () ->
                                                        let current_result = recreate_expression_from_b_values b_values in
                                                        if arg.generate_automata then regenerate_automata b_values;
                                                        current_result)
                                 end
                                 else
                                 begin
                                         if arg.verbose then arg.print "Since the minimum invariant is not needed, computing stops.\n\n";
                                         if arg.generate_automata then regenerate_automata b_values;
                                         recreate_expression_from_model (get_model context_H)
                                 end
                 | L_TRUE ->
                                 let sigma = get_model context_F in
                                 if arg.verbose then
                                 begin
                                         arg.print "This assignment for B does not contain all the constraint it is wanted to do.\n";
                                         arg.print "Here is an assignement for the variable that shows it :\n";
                                         print_model_F arg sigma;
                                         arg.print "Adding contraints on the system and retrying.\n\n"
                                 end;
                                 assert_expr context_H (create_assert_F_Sigma_sigma sigma);
                                 pop context_F; solve arg last_result

let get_automata () =
        automata

let reset () =
        pop context_F;
        reset context_F;
        pop context_H;
        reset context_H;
        Alpha.reset_yices_var ();
        ()

