(*---------------------------------------------------
--
-- Typage des expressions Lustre
--
----------------------------------------------------*)

open Hashtbl
open XecErrors
open Lexeme
open Syntaxe
open Printf


(* un type_eff passe  travers les aliases *)
type type_eff =
	TE_bool
|	TE_int
|	TE_real
|	TE_extern of string
|	TE_enum of (string * string list)
|	TE_struct of (string * (string * type_eff) list)
|	TE_array of (type_eff * int)

let rec string_of_type_eff = function
   TE_bool -> "bool"
|  TE_int  -> "int"
|  TE_real -> "real"
|  TE_extern (s) -> s^"(*extern*)"
|  TE_enum (s, sl) -> ignore sl; s^"(*enum*)"
|  TE_struct (s, fl) -> ignore fl; s^"(*sruct*)"
|  TE_array (ty, sz) -> Printf.sprintf "%s^%d" (string_of_type_eff ty) sz
and string_of_type_eff_list = function
   []  -> ""
|  [x] -> (string_of_type_eff x)
|  l   -> (
   String.concat " * " (List.map string_of_type_eff l)
)


type profile = (type_eff list * type_eff list)
type env = {
	type_tab : (string, type_eff option) Hashtbl.t ;
	flow_type_tab : (string, type_eff) Hashtbl.t ;
	flow_clock_tab : (string, clock_exp) Hashtbl.t ;
	oper_tab : (string, profile) Hashtbl.t ;
	(*exp_tab : (Syntaxe.val_exp, type_eff list) Hashtbl.t ;*)
}

(* l'env bidon vide *)
let empty_env = {
	type_tab = Hashtbl.create 50;
	flow_type_tab = Hashtbl.create 50;
	flow_clock_tab = Hashtbl.create 50;
	oper_tab = Hashtbl.create 50;
}

let eval_const_int (ctx : env) (e : Syntaxe.val_exp) = ignore ctx; (
(* pas de fioriture pour l'instant => il faut une constante *)
	match e with
		(ICONST_n s, Ops []) -> int_of_ecint s
	| _ -> raise (Global_error "eval_const_int: unexpected case") 
)

(* utilitaire *)
let snd_of_3 = function (_, b, _) -> b

(*---------------------------------------------------
FONCTION : type_exp_of_type_eff 
-----------------------------------------------------
(type_eff -> type_exp)
Utile pour re-gnrer du Lustre correct


N.B. A FAIRE pour l'instant on ne tient pas compte
des ventuels alias dans le ctx ... a voir ...
----------------------------------------------------*)
let rec type_exp_of_type_eff (ctx : env) (ty : type_eff) = (
	match ty with
		TE_bool -> Bool
	|	TE_int  -> Int
	|	TE_real -> Real
	|	TE_extern s -> NamedType s
	|	TE_enum (s, _) -> NamedType s
	|	TE_struct (s, _) -> NamedType s
	|	TE_array (telt, n) -> (
		let telt_exp = type_exp_of_type_eff ctx telt in
		let nexp = Syntaxe.int_exp n in
		ArrayType (telt_exp, nexp)
	)
)


(*---------------------------------------------------
FONCTION : ident_type
(string -> type_eff)
----------------------------------------------------*)
let ident_type (ctx : env) (s : string) = (
	try (
		Hashtbl.find  ctx.flow_type_tab s
	) with Not_found -> (
		raise(Global_error("Typing.ident_type: undeclared flow: "^s))
	)
)

(*---------------------------------------------------
-- FONCTION PRINCIPALE : exp_type
--
-- De manire gnrale, une expression peut tre un tuple,
-- (appels rendant plusieurs rsultats) c'est pourquoi on 
-- renvoie une liste de type_eff
----------------------------------------------------*)
exception Type_error of string 
let rec exp_type (ctx : env) (e : Syntaxe.val_exp) = (
	(* utilitaire : liste de bool ? *)
	let rec narybool = ( function
			[TE_bool] -> [TE_bool]
		|	TE_bool::tail -> (narybool tail)
		|	_ -> raise(Type_error("bad type combination"))
	) in 
	(* utilitaire : quand la rgle est trs prcise ... *)
	let simprule expected result given = (
		if ( expected = given) then [result]
		else raise(Type_error("bad type combination"))
	) in
	(* utilitaire : polymorphe ... *)
	let poly1 = ( function
		[te] -> [te]
		| _ -> raise(Type_error("bad type combination"))
	) in
	let poly2 = ( function
		[te1; te2] -> if (te1 = te2) then [te1]
		              else raise(Type_error("bad type combination"))
		| _ -> raise(Type_error("bad type combination"))
	) in
	let polybool = ( function
		[te1; te2] -> if (te1 = te2) then [TE_bool]
		              else raise(Type_error("bad type combination"))
		| _ -> raise(Type_error("bad type combination"))
	) in
	match e with (nat, Ops ops) ->
	(* On calcule d'ores et dja la liste les types des arguments *)
	let tops = List.flatten (List.map (exp_type ctx) ops) in
	match nat with
	(* feuilles *)
   	TRUE_n     -> [TE_bool]	
	|  FALSE_n    -> [TE_bool]
        |  ICONST_n s -> ignore s; [TE_int]
        |  RCONST_n s -> ignore s; [TE_real]
        |  IVAL_n   i -> ignore i; [TE_int]
        |  RVAL_n   r -> ignore r; [TE_real]
	|  IDENT_n  s -> [ident_type ctx s]
	|	NOT_n      -> (simprule [TE_bool] TE_bool tops)
	|	UMINUS_n   -> (try (simprule [TE_int] TE_int tops)
	                 with Type_error _ ->
	                     (simprule [TE_real] TE_real tops)
	)
	|	PRE_n 
	|	CURRENT_n  -> (poly1 tops)
	|	INT2REAL_n -> (simprule [TE_int] TE_real tops)
	|	REAL2INT_n -> (simprule [TE_real] TE_int tops)
	|	ARROW_n    -> (poly2 tops)
	|	AND_n
	|	OR_n
	|	XOR_n
	|	IMPL_n     -> (simprule [TE_bool; TE_bool] TE_bool tops)
	|	EQ_n
	|	NEQ_n      -> (polybool tops)
	|  GTE_n
	|  GT_n
	|  LTE_n
	|  LT_n       -> (try (simprule [TE_int;TE_int] TE_bool tops)
	                 with Type_error _ ->
	                     (simprule [TE_real;TE_real] TE_bool tops)
	)
	|  MINUS_n
	|  PLUS_n
	|  SLASH_n
	|  TIMES_n    -> (try (simprule [TE_int;TE_int] TE_int tops)
	                 with Type_error _ ->
	                     (simprule [TE_real;TE_real] TE_real tops)
	)
	|  DIV_n
	|  MOD_n      -> (simprule [TE_int;TE_int] TE_int tops)
	|  IF_n       -> (
	                 match tops with TE_bool::tail -> (poly2 tail)
	                 | _ -> raise(Type_error("bad type combination")) 
	)
	|	NOR_n
	|	DIESE_n    -> (narybool tops) 
(* speciaux *)
	|  TUPLE_n    -> (tops)
	|  CALL_n f   -> (
	                 let (pri, pro) = ( try (Hashtbl.find ctx.oper_tab f)
	                 with Not_found ->
	                   raise(Global_error("undeclared operator"^f)) 
	                 ) in 
	                 if (pri = tops) then pro
	                 else raise(Type_error("bad type combination")) 
	)
(*
|  HAT_n
|  ARRAY_n
|  ARRAY_ACCES_n i
|  STRUCT_ACCESS_n s
*)
	(* sinon erreur ... *)
	| _ -> raise(Internal_error ("Typing.exp_type","operator not implemented"))
)

(*---------------------------------------------------
-- FONCTION get_env
--
-- Construction de l'environnement :
--	. on parcours le package pour rcolter les constantes
--	et let type_eff des types dclars 
----------------------------------------------------*)
(* ajout d'une entre dans flow_type_tab *)
let declare_flow
(this : env ) (t : type_eff) (c : clock_exp) (flow : string) = (
	try (
		let _ = Hashtbl.find this.flow_type_tab flow in
		raise (Global_error ("flow declared twice: "^flow))
	) with Not_found -> (
		Hashtbl.add this.flow_type_tab flow t;
		Hashtbl.add this.flow_clock_tab flow c
	)
)

	
let get_env (p : Syntaxe.package) (nopt : Syntaxe.node_info option) = (
	let this = {
		type_tab = Hashtbl.create 50;
		flow_type_tab = Hashtbl.create 50;
		flow_clock_tab = Hashtbl.create 50;
		oper_tab = Hashtbl.create 50;
		(*exp_tab = Hashtbl.create 50;*)
	} in

	(* ajout d'un (profile, id) dans oper_tab *)
	let add_oper_type (t : profile) (op : string) = (
		try (
			let _ = Hashtbl.find this.oper_tab op in
			raise (Global_error ("operator declared twice: "^op))
		) with Not_found -> (
			Hashtbl.add this.oper_tab op t
		)
	) in
	(*************************************************************************)
	(* traitement des types                                                  *)
	(*************************************************************************)
	let rec type_eff_of_type_exp (texp : Syntaxe.type_exp) = (
		match texp with
				Bool -> TE_bool
			|	Int -> TE_int
			|	Real -> TE_real
			|	NamedType id -> type_eff_of_type_id id
			|	ArrayType (te, ev) -> (
				let sz = eval_const_int this ev in
				let t = type_eff_of_type_exp te in  
				TE_array (t, sz)
			)
			|	TupleType _ -> raise (Global_error "tuple types not supported")
	) and type_eff_of_type_id ( id : string ) = (
		(* dj vu ? *)
		try (
			match (Hashtbl.find this.type_tab id) with
				Some t -> t
			|	None -> raise (Global_error
				("recursive definition of type "^id))	
		) with Not_found -> (
			(* pas encore vu ... *)
			Hashtbl.add this.type_tab id None ;
			let ti_flg = try (
				Hashtbl.find (p.type_table) id
			) with Not_found ->
				raise (Global_error ("undeclared type "^id)) in
			let res = type_eff_of_type_info ti_flg in
			Hashtbl.add this.type_tab id (Some res) ;
			res
		)		
	) and type_eff_of_type_info = ( function 
                { it = tinfo ; src = lxm } -> ignore lxm; (
			match tinfo with
			ExternalType id -> (
				TE_extern id
			) |
			AliasedType (id, te) -> (
                                ignore id; type_eff_of_type_exp te	
			) |
			EnumType (id, vals) -> (
				(* on mets toutes les constantes dans la table *)
				let res = TE_enum (id, vals) in
				List.iter (declare_flow this res BaseClock) vals ;
				res
			) |
			StructType (id, flds) -> (
				let field_eff = function
					(f, te) -> (f, type_eff_of_type_exp te) 
				in TE_struct (id, (List.map field_eff flds))
			)
		)
	) in
	(* on itere sur la table des dclarations ... *)	
        let treat_type_decl x ti = ignore x; (
		let _ = type_eff_of_type_info ti in ()
	) in
	Hashtbl.iter treat_type_decl p.type_table  ;

	(*************************************************************************)
	(* traitement des constantes                                             *)
	(*************************************************************************)
	let treat_const_decl _ ci_flg = (
		match ci_flg.it with
		ExternalConst (id, te) -> (
			declare_flow this (type_eff_of_type_exp te) BaseClock id
		) |
		DefinedConst (id, Some te, _) -> (
			declare_flow this (type_eff_of_type_exp te) BaseClock id
		) |
		DefinedConst (id, None, e) -> (
			match (exp_type this e) with
				[te] -> declare_flow this te BaseClock id
			| _ -> raise (Global_error ("tuple const not supported"))
		)
	) in 	
	Hashtbl.iter treat_const_decl p.const_table  ;
	(*************************************************************************)
	(* traitement des variables du node                                      *)
	(*************************************************************************)
	( match (nopt) with 
		Some n -> (
			let treat_var_decl = function (id, te, ce) -> (
				declare_flow this (type_eff_of_type_exp te) ce id	
			) in
			List.iter treat_var_decl n.ni_inputs ;
			List.iter treat_var_decl n.ni_outputs ;
			List.iter treat_var_decl n.ni_locals ;
		) | _ -> ()
	);
	(*************************************************************************)
	(* traitement des operateurs                                             *)
	(*************************************************************************)
	let treat_oper_decl id oi_flg = (
		match oi_flg.it with
		FuncOper fi -> (
			add_oper_type (
				List.map type_eff_of_type_exp (List.map snd fi.fi_inpars),
				List.map type_eff_of_type_exp (List.map snd fi.fi_outpars)
			) id
		) |
		NodeOper ni -> (
			add_oper_type (
				List.map type_eff_of_type_exp (List.map snd_of_3 ni.ni_inputs),
				List.map type_eff_of_type_exp (List.map snd_of_3 ni.ni_outputs)
			) id
		)
	) in
	Hashtbl.iter treat_oper_decl p.oper_table  ;
	(*************************************************************************)
	(* fini, on renvoie this ...                                             *)
	(*************************************************************************)
	this
)


(* Iteration sur l'ensemble des types nomms *)

let iter_on_types (this : env) (proc : (string -> type_eff -> unit)) = (
	(* capte le flag option *)
	let correctproc id teopt = (
		match teopt with
			Some te -> (proc id te)
		|	None -> () 
	) in 
	Hashtbl.iter correctproc this.type_tab 
)


(**************************************************************************)
(*            CLOCKING                                                    *)
(**************************************************************************)

(*---------------------------------------------------
-- FONCTION PRINCIPALE : exp_clock
--
-- Restriction : les expressions, mme tuple,
-- doivent tre mono-clock
----------------------------------------------------*)
exception Clock_error of string 

let rec exp_clock (ctx : env) (e : Syntaxe.val_exp) = (
	match e with (nat, Ops ops) ->
	(* On calcule d'ores et dja la liste des clocks des arguments *)
	let cops = List.map (exp_clock ctx) ops in
	match nat with
	(* constantes => base clock *)
   	TRUE_n     
	|  FALSE_n    
	|  ICONST_n _ 
	|  RCONST_n _ 
	|  IVAL_n   _ 
	|  RVAL_n   _ -> BaseClock
	(* ident => voir ctx *)
	|  IDENT_n  s -> ident_clock ctx s
	(* current *)
	|	CURRENT_n  -> (
		match cops with
			(NamedClock s)::[] -> (
				ident_clock ctx s
			) |
			_ -> (
				raise (Global_error "exp_clock : unexpected case in CURRENT_n")
			)
	)
	(* when *)
	|	WHEN_n  -> (
		match cops with
			c0::c1::[] -> (
				if (c0 = c1) then (
					match (List.nth ops 1) with
						(IDENT_n s , _ ) -> NamedClock s
					|	_ -> (
						raise (Clock_error("snd operand of when must be a var"))
					) 
				) else (
					raise (Clock_error("operands not on same clock"))
				)
			)
			| _ -> (
				raise (Global_error "exp_clock : unexpected case in WHEN_n")
			)
	)
	(* cas gnral *)
	|	_ -> (
		match cops with
			c0 :: t -> (
				let check_clock c = (
					if (c = c0) then ()
					else (
						raise (Clock_error("operands not on same clock"))
					)
				) in
				List.iter check_clock t ;
				c0
			)
		|	_ -> (
			raise (Global_error "exp_clock : unexpected case")
		)
	)
) and ident_clock (ctx : env) (s : string) = (
	try (
		Hashtbl.find  ctx.flow_clock_tab s
	) with Not_found -> (
		raise(Global_error("exp_clock:  undeclared flow: "^s))
	)
)
